import java.io.EOFException;
import java.io.PrintStream;
import java.util.*;

class CoursePlanner
{
  private AdjList prereqGraph; // Courses are verticies, prereqs as edges
  private String courseName[]; // Name to number mapping

  /*
   * A basic copy constructor
   */
  public CoursePlanner(CoursePlanner courseplanner)
  {
    prereqGraph = new AdjList(courseplanner.prereqGraph);
    courseName = (String[])courseplanner.courseName.clone();
  }


  /*
   * Constructor to create the CoursePlanner from a file
   * containing the course list and dependency information
   */
  public CoursePlanner(String file)
  {
    DataReader datareader = new DataReader(); 

    try
    {
      datareader = new DataReader(file);

      // Read the number of verticies and creata data structures
      int numVerticies = datareader.readInt();
      courseName = new String[numVerticies];
      prereqGraph = new AdjList(numVerticies);

      // Build map form names to numbers
      for(int vertexNumber = 0; vertexNumber < numVerticies; vertexNumber++)
      {
        courseName[vertexNumber] = datareader.readWord();
      }

    }
    catch(Exception e)
    {
      // A more-or-less fatal problem
      System.out.println(e);
      System.exit(-1);
    }


    // Build graph from dependencies
    try
    {
      while (true)
      {
        String courseName = datareader.readWord();
        String prereqName = datareader.readWord();
        prereqGraph.addEdge(new Edge(getCourseNumber(courseName), 
                                     getCourseNumber(prereqName)));
      }
    }

    catch(EOFException eof) 
    { 
      // This is a normal exit
    }

    catch(Exception e)
    {
      // This is a problem. What to do? Exit
      System.out.println(e);
      System.exit(-1);
    }
  }


  /*
   * Given a course name, return the number
   * the reverse is done by indexing directly into the vector
   */
  private int getCourseNumber(String s)
  {
    for(int nameIndex = 0; nameIndex < courseName.length; nameIndex++)
      if(courseName[nameIndex].equals(s))
        return nameIndex;

    // Couldn't find a match
    return -1;
  }

  
  /*
   * This is the simple program that prodces a valid schedule given
   * an input file
   */
  public static void main(String args[])
  {
      if(args.length==0){
	  System.err.println("Please specify a data file as a command-line argument");
	  System.exit(-1);
      }
    CoursePlanner courseplanner = new CoursePlanner(args[0]);
    System.out.println("The following are the dependencies:\n");
    System.out.println(courseplanner); 

    try
    {
      String s = courseplanner.proposeSchedule().toString();
      System.out.println("The following is one possible schedule:\n");
      System.out.println("\t" + s);
    }
    catch(AdjList.UnsatisfiedException use)
    {
      System.out.println("The following prevent a valid schedule:\n");
      System.out.println(courseplanner);
    }
  }


  /*
   * This method invoces the topologicalSort() method of the
   * AdjList class to provide a valid schedule, given the
   * constraints
   *
   * As supported by topologicalSort(), it should either print
   * out the valid schedule, or print out an error message
   * and the unsatisfied dependencies (verticies involved in cycles).
   */
  public LinkedList proposeSchedule() throws AdjList.UnsatisfiedException
  {
    LinkedList scheduleByName = new LinkedList();
    int i;

    LinkedList scheduleByNumber = prereqGraph.topologicalSort(); 

    while (scheduleByNumber.size() !=0)
    {
        int courseNumber = ((Integer)scheduleByNumber.removeFirst()).intValue();
        scheduleByName.addLast(courseName[courseNumber]);
    }

    return scheduleByName;
  }


  /*
   * A fairly ugly but functional way to represent a CoursePlanner as
   * a String
   */
  public String toString()
  {
     String ret = "";

     for(int courseNumber = 0; courseNumber < courseName.length; courseNumber++)
     {
       String prereqs = "";
       ListIterator listiterator = prereqGraph.getAdjacencies(courseNumber);

       while(listiterator.hasNext())
	   prereqs += "\t" + 
	       courseName[((Edge)listiterator.next()).getDestinationVertex()] 
	       + " ";

       ret+="\t"+courseNumber+": "+courseName[courseNumber]+" requires ";
       if(prereqs != "")
	   ret+=prereqs;
       else
	   ret+="nothing";
       ret+="\n";
     }

     return ret;
  }
}
